@@ -28,21 +28,19 @@
  *     MA  02110-1301  USA
  */
 
-#define NUMUDPARGS       2
+#define NUMUDPARGS       1
 #define NUMUDPINPUTBODYS 1
 #include "udpUtilities.h"
 
 /* shorthands for accessing argument values and velocities */
 #define EDGELIST(    IUDP,I)  ((int    *) (udps[IUDP].arg[0].val))[I]
 #define EDGELIST_SIZ(IUDP)    ( int     )  udps[IUDP].arg[0].size
-#define NSMTH(       IUDP)    ((int    *) (udps[IUDP].arg[1].val))[0]
-#define NSMTH_SIZ(   IUDP)    ( int     )  udps[IUDP].arg[1].size
 
 /* data about possible arguments */
-static char  *argNames[NUMUDPARGS] = {"edgelist", "nsmth", };
-static int    argTypes[NUMUDPARGS] = {ATTRINT,    ATTRINT, };
-static int    argIdefs[NUMUDPARGS] = {0,          0,       };
-static double argDdefs[NUMUDPARGS] = {0.,         0.,      };
+static char  *argNames[NUMUDPARGS] = {"edgelist", };
+static int    argTypes[NUMUDPARGS] = {ATTRINT,    };
+static int    argIdefs[NUMUDPARGS] = {0,          };
+static double argDdefs[NUMUDPARGS] = {0.,         };
 
 /* get utility routines: udpErrorStr, udpInitialize, udpReset, udpSet,
                          udpGet, udpVel, udpClean, udpMesh */
@@ -72,9 +70,11 @@ udpExecute(ego  emodel,                 /* (in)  Model containing Body */
     int     oclass, mtype, nchild, nchild2, *senses, *senses2, i, im1, j, foo;
     int     nedge, myNedge, ibeg, ncorn=0, one=1, ntemp, *stemps, attrType, attrLen;
     int     *header_s=NULL, *header_e=NULL, *header_n=NULL, *header_w=NULL, *header=NULL;
-    int     periodic, imax, jmax, ismth;
+    int     periodic, imax, jmax;
     CINT    *tempIlist;
-    double  trange[2], limits[4], xyz[18], *xyztemp=NULL, omega=1.0;
+    double  trange[2], limits[4], xyz[18];
+    double  fraci, fracj, H0i, H1i, H2i, H3i, H0j, H1j, H2j, H3j, bcwe, bcsn;
+    double  *lens=NULL, *lenn=NULL, *lenw=NULL, *lene=NULL;
     double  data[4], dataim1[18], datai[18], *dotprod=NULL;
     double  *rdata_s=NULL, *rdata_e=NULL, *rdata_n=NULL, *rdata_w=NULL, *rdata=NULL;
     CDOUBLE *tempRlist;
@@ -95,7 +95,6 @@ udpExecute(ego  emodel,                 /* (in)  Model containing Body */
     for (i = 0; i < EDGELIST_SIZ(0); i++) {
         printf("edgelist[%2d] = %d\n", i, EDGELIST(0,i));
     }
-    printf("nsmth = %d\n", NSMTH(0));
 #endif
 
     /* default return values */
@@ -132,12 +131,6 @@ udpExecute(ego  emodel,                 /* (in)  Model containing Body */
     echilds = eedges;                   // needed to avoid a clang warning
 
     /* check arguments */
-    if (NSMTH_SIZ(0) > 1) {
-        snprintf(message, 1024, "\"nsmth\" should be a scalar");
-        status = EGADS_RANGERR;
-        goto cleanup;
-    }
-
     for (i = 0; i < EDGELIST_SIZ(0); i++) {
         if (EDGELIST(0,i) < 1 || EDGELIST(0,i) > nedge) {
             snprintf(message, 1024, "edgelist[%d]=%d is out of bounds", i, EDGELIST(0,i));
@@ -154,7 +147,6 @@ udpExecute(ego  emodel,                 /* (in)  Model containing Body */
     for (i = 0; i < EDGELIST_SIZ(0); i++) {
         printf("edgelist[%2d,%2d] = %d\n", i, numUdp, EDGELIST(numUdp,i));
     }
-    printf("nsmth[      %2d] = %d\n", numUdp, NSMTH(numUdp));
 #endif
 
     /* make a Loop of the Edges in edgelist */
@@ -185,7 +177,6 @@ udpExecute(ego  emodel,                 /* (in)  Model containing Body */
 
     /* we have to start the Loop at a corner */
     for (; ibeg < EDGELIST_SIZ(0); ibeg++) {
-
         myNedge = 0;
         for (i = 0; i < EDGELIST_SIZ(0); i++) {
             j = (i+ibeg) % EDGELIST_SIZ(0);
@@ -320,12 +311,23 @@ udpExecute(ego  emodel,                 /* (in)  Model containing Body */
     }
 #endif
 
+    /* bc types in TFI below is set if any Edges have the _patchBC attribute */
+    bcwe = 0;
+    bcsn = 0;
+
     /* create the Bsplines for the 4 sides of the patch */
     i = 0;
 
     /* south side */
     status = EG_convertToBSpline(echilds[i], &bspline_s);
     CHECK_STATUS(EG_convertToBSpline);
+
+    status = EG_attributeRet(echilds[i], "_patchBC", &attrType, &attrLen,
+                             &tempIlist, &tempRlist, &tempClist);
+    if (status == EGADS_SUCCESS && attrType == ATTRREAL) {
+        bcsn = tempRlist[0];
+    }
+
     i++;
 
     for (; i < nchild; i++) {
@@ -335,6 +337,12 @@ udpExecute(ego  emodel,                 /* (in)  Model containing Body */
             status = EG_convertToBSpline(echilds[i], &bspline_new);
             CHECK_STATUS(EG_convertToBSpline);
 
+            status = EG_attributeRet(echilds[i], "_patchBC", &attrType, &attrLen,
+                                     &tempIlist, &tempRlist, &tempClist);
+            if (status == EGADS_SUCCESS && attrType == ATTRREAL) {
+                bcsn = tempRlist[0];
+            }
+
             SPLINT_CHECK_FOR_NULL(bspline_old);
             SPLINT_CHECK_FOR_NULL(bspline_new);
 
@@ -349,6 +357,13 @@ udpExecute(ego  emodel,                 /* (in)  Model containing Body */
         } else {
             status = EG_convertToBSpline(echilds[i], &bspline_e);
             CHECK_STATUS(EG_convertToBSpline);
+
+            status = EG_attributeRet(echilds[i], "_patchBC", &attrType, &attrLen,
+                                     &tempIlist, &tempRlist, &tempClist);
+            if (status == EGADS_SUCCESS && attrType == ATTRREAL) {
+                bcwe = tempRlist[0];
+            }
+
             i++;
 
             break;
@@ -363,6 +378,12 @@ udpExecute(ego  emodel,                 /* (in)  Model containing Body */
             status = EG_convertToBSpline(echilds[i], &bspline_new);
             CHECK_STATUS(EG_convertToBSpline);
 
+            status = EG_attributeRet(echilds[i], "_patchBC", &attrType, &attrLen,
+                                     &tempIlist, &tempRlist, &tempClist);
+            if (status == EGADS_SUCCESS && attrType == ATTRREAL) {
+                bcwe = tempRlist[0];
+            }
+
             SPLINT_CHECK_FOR_NULL(bspline_old);
             SPLINT_CHECK_FOR_NULL(bspline_new);
 
@@ -377,6 +398,13 @@ udpExecute(ego  emodel,                 /* (in)  Model containing Body */
         } else {
             status = EG_convertToBSpline(echilds[i], &bspline_n);
             CHECK_STATUS(EG_convertToBSpline);
+
+            status = EG_attributeRet(echilds[i], "_patchBC", &attrType, &attrLen,
+                                     &tempIlist, &tempRlist, &tempClist);
+            if (status == EGADS_SUCCESS && attrType == ATTRREAL) {
+                bcsn = tempRlist[0];
+            }
+
             i++;
 
             break;
@@ -391,6 +419,12 @@ udpExecute(ego  emodel,                 /* (in)  Model containing Body */
             status = EG_convertToBSpline(echilds[i], &bspline_new);
             CHECK_STATUS(EG_convertToBSpline);
 
+            status = EG_attributeRet(echilds[i], "_patchBC", &attrType, &attrLen,
+                                     &tempIlist, &tempRlist, &tempClist);
+            if (status == EGADS_SUCCESS && attrType == ATTRREAL) {
+                bcsn = tempRlist[0];
+            }
+
             SPLINT_CHECK_FOR_NULL(bspline_old);
             SPLINT_CHECK_FOR_NULL(bspline_new);
 
@@ -406,6 +440,13 @@ udpExecute(ego  emodel,                 /* (in)  Model containing Body */
             if (ncorn == 4) {
                 status = EG_convertToBSpline(echilds[i], &bspline_w);
                 CHECK_STATUS(EG_convertToBSpline);
+
+                status = EG_attributeRet(echilds[i], "_patchBC", &attrType, &attrLen,
+                                         &tempIlist, &tempRlist, &tempClist);
+                if (status == EGADS_SUCCESS && attrType == ATTRREAL) {
+                    bcwe = tempRlist[0];
+                }
+
                 i++;
             }
 
@@ -423,6 +464,12 @@ udpExecute(ego  emodel,                 /* (in)  Model containing Body */
             status = EG_convertToBSpline(echilds[i], &bspline_new);
             CHECK_STATUS(EG_convertToBSpline);
 
+            status = EG_attributeRet(echilds[i], "_patchBC", &attrType, &attrLen,
+                                     &tempIlist, &tempRlist, &tempClist);
+            if (status == EGADS_SUCCESS && attrType == ATTRREAL) {
+                bcwe = tempRlist[0];
+            }
+
             SPLINT_CHECK_FOR_NULL(bspline_old);
             SPLINT_CHECK_FOR_NULL(bspline_new);
 
@@ -446,7 +493,7 @@ udpExecute(ego  emodel,                 /* (in)  Model containing Body */
     CHECK_STATUS(EG_getGeometry);
 
 #ifdef DEBUG
-    printf("after merging: south\n");
+    printf("after merging: south (bcsn=%10.5f)\n", bcsn);
     ocsmPrintEgo(bspline_s);
 #endif
 
@@ -456,7 +503,7 @@ udpExecute(ego  emodel,                 /* (in)  Model containing Body */
     CHECK_STATUS(EG_getGeometry);
 
 #ifdef DEBUG
-    printf("after merging: west\n");
+    printf("after merging: west  (bcwe=%10.5f)\n", bcwe);
     ocsmPrintEgo(bspline_w);
 #endif
 
@@ -466,7 +513,7 @@ udpExecute(ego  emodel,                 /* (in)  Model containing Body */
     CHECK_STATUS(EG_getGeometry);
 
 #ifdef DEBUG
-    printf("after merging: north\n");
+    printf("after merging: north (bcsn=%10.5f)\n", bcsn);
     ocsmPrintEgo(bspline_n);
 #endif
 
@@ -477,7 +524,7 @@ udpExecute(ego  emodel,                 /* (in)  Model containing Body */
         CHECK_STATUS(EG_getGeometry);
 
 #ifdef DEBUG
-        printf("after merging: west\n");
+        printf("after merging: west  (bcwe=%10.5f)\n", bcwe);
         ocsmPrintEgo(bspline_w);
 #endif
     }
@@ -600,7 +647,6 @@ udpExecute(ego  emodel,                 /* (in)  Model containing Body */
     status = EG_isoCline(newLoop, 0, 0, &newSurface);
     CHECK_STATUS(EG_isoCline);
 
-    /* smooth the control points if required */
     if (header != NULL) {EG_free(header);   header = NULL;}
     if (rdata  != NULL) {EG_free(rdata );   rdata  = NULL;}
 
@@ -614,38 +660,133 @@ udpExecute(ego  emodel,                 /* (in)  Model containing Body */
     imax = header[2];
     jmax = header[5];
 
-    MALLOC(xyztemp, double, header[3]+header[6]+4*imax*jmax);
-
 #define XDATA(I,J)  rdata[header[3]+header[6]+3*((I)+(J)*imax)  ]
 #define YDATA(I,J)  rdata[header[3]+header[6]+3*((I)+(J)*imax)+1]
 #define ZDATA(I,J)  rdata[header[3]+header[6]+3*((I)+(J)*imax)+2]
 
-#define XTEMP(I,J)  xyztemp[header[3]+header[6]+3*((I)+(J)*imax)  ]
-#define YTEMP(I,J)  xyztemp[header[3]+header[6]+3*((I)+(J)*imax)+1]
-#define ZTEMP(I,J)  xyztemp[header[3]+header[6]+3*((I)+(J)*imax)+2]
+    /* slopes at the boundaries (+1 for slope set, -1 for natural) */
+#define DXS(I)  (XDATA(I,jmax-1) - XDATA(I,0)) / (jmax - 1)
+#define DYS(I)  (YDATA(I,jmax-1) - YDATA(I,0)) / (jmax - 1)
+#define DZS(I)  (ZDATA(I,jmax-1) - ZDATA(I,0)) / (jmax - 1)
+#define DXN(I)  (XDATA(I,jmax-1) - XDATA(I,0)) / (jmax - 1)
+#define DYN(I)  (YDATA(I,jmax-1) - YDATA(I,0)) / (jmax - 1)
+#define DZN(I)  (ZDATA(I,jmax-1) - ZDATA(I,0)) / (jmax - 1)
+#define DXW(J)  (XDATA(imax-1,J) - XDATA(0,J)) / (imax - 1)
+#define DYW(J)  (YDATA(imax-1,J) - YDATA(0,J)) / (imax - 1)
+#define DZW(J)  (ZDATA(imax-1,J) - ZDATA(0,J)) / (imax - 1)
+#define DXE(J)  (XDATA(imax-1,J) - XDATA(0,J)) / (imax - 1)
+#define DYE(J)  (YDATA(imax-1,J) - YDATA(0,J)) / (imax - 1)
+#define DZE(J)  (ZDATA(imax-1,J) - ZDATA(0,J)) / (imax - 1)
 
-    for (ismth = 0; ismth < NSMTH(0); ismth++) {
-        for (j = 1; j < jmax-1; j++) {
-            for (i = 1; i < imax-1; i++) {
-                XTEMP(i,j) = ( XDATA(i-1,j  ) + XDATA(i+1,j  )
-                             + XDATA(i  ,j-1) + XDATA(i  ,j+1)) / 4;
-                YTEMP(i,j) = ( YDATA(i-1,j  ) + YDATA(i+1,j  )
-                             + YDATA(i  ,j-1) + YDATA(i  ,j+1)) / 4;
-                ZTEMP(i,j) = ( ZDATA(i-1,j  ) + ZDATA(i+1,j  )
-                             + ZDATA(i  ,j-1) + ZDATA(i  ,j+1)) / 4;
-            }
+    /* determine the arclengths along the boundaries */
+    MALLOC(lens, double, imax);
+    MALLOC(lenn, double, imax);
+    MALLOC(lenw, double, jmax);
+    MALLOC(lene, double, jmax);
+
+    lens[0] = 0;
+    lenn[0] = 0;
+    for (i = 1; i < imax; i++) {
+        j = 0;
+        lens[i] = lens[i-1] + sqrt((XDATA(i,j)-XDATA(i-1,j))*(XDATA(i,j)-XDATA(i-1,j))
+                                  +(YDATA(i,j)-YDATA(i-1,j))*(YDATA(i,j)-YDATA(i-1,j))
+                                  +(ZDATA(i,j)-ZDATA(i-1,j))*(ZDATA(i,j)-ZDATA(i-1,j)));
+        j = jmax-1;
+        lenn[i] = lenn[i-1] + sqrt((XDATA(i,j)-XDATA(i-1,j))*(XDATA(i,j)-XDATA(i-1,j))
+                                  +(YDATA(i,j)-YDATA(i-1,j))*(YDATA(i,j)-YDATA(i-1,j))
+                                  +(ZDATA(i,j)-ZDATA(i-1,j))*(ZDATA(i,j)-ZDATA(i-1,j)));
+    }
+
+    lenw[0] = 0;
+    lene[0] = 0;
+    for (j = 1; j < jmax; j++) {
+        i = 0;
+        lenw[j] = lenw[j-1] + sqrt((XDATA(i,j)-XDATA(i,j-1))*(XDATA(i,j)-XDATA(i,j-1))
+                                  +(YDATA(i,j)-YDATA(i,j-1))*(YDATA(i,j)-YDATA(i,j-1))
+                                  +(ZDATA(i,j)-ZDATA(i,j-1))*(ZDATA(i,j)-ZDATA(i,j-1)));
+        i = imax-1;
+        lene[j] = lene[j-1] + sqrt((XDATA(i,j)-XDATA(i,j-1))*(XDATA(i,j)-XDATA(i,j-1))
+                                  +(YDATA(i,j)-YDATA(i,j-1))*(YDATA(i,j)-YDATA(i,j-1))
+                                  +(ZDATA(i,j)-ZDATA(i,j-1))*(ZDATA(i,j)-ZDATA(i,j-1)));
+    }
+
+    /* update the interior control points using arc-length-weighted cubic TFI */
+    for (j = 1; j < jmax-1; j++) {
+        fracj = (lenw[j] / lenw[jmax-1] + lene[j] / lene[jmax-1]) / 2;
+
+        if (bcsn != 0) {
+            H0j = (fracj * (fracj * ( 2 * fracj - 3)    ) + 1);
+            H1j = (fracj * (fracj * ( 1 * fracj - 2) + 1)    ) / bcsn;
+            H2j = (fracj * (fracj * ( 1 * fracj - 1)    )    ) / bcsn;
+            H3j = (fracj * (fracj * (-2 * fracj + 3)    )    );
+        } else {
+            H0j = (1 - fracj);
+            H1j = 0;
+            H2j = 0;
+            H3j =      fracj;
         }
 
-        for (j = 1; j < jmax-1; j++) {
-            for (i = 1; i < imax-1; i++) {
-                XDATA(i,j) = XDATA(i,j) + omega * (XTEMP(i,j) - XDATA(i,j));
-                YDATA(i,j) = YDATA(i,j) + omega * (YTEMP(i,j) - YDATA(i,j));
-                ZDATA(i,j) = ZDATA(i,j) + omega * (ZTEMP(i,j) - ZDATA(i,j));
+        for (i = 1; i < imax-1; i++) {
+            fraci = (lens[i] / lens[imax-1] + lenn[i] / lenn[imax-1]) / 2;
+
+            if (bcwe != 0) {
+                H0i = (fraci * (fraci * ( 2 * fraci - 3)    ) + 1);
+                H1i = (fraci * (fraci * ( 1 * fraci - 2) + 1)    ) / bcwe;
+                H2i = (fraci * (fraci * ( 1 * fraci - 1)    )    ) / bcwe;
+                H3i = (fraci * (fraci * (-2 * fraci + 3)    )    );
+            } else {
+                H0i = (1 - fraci);
+                H1i = 0;
+                H2i = 0;
+                H3i =      fraci;
             }
+
+            XDATA(i,j) =
+                + H0i       * XDATA(0     ,j     )
+                + H1i       * DXW(         j     )
+                + H2i       * DXE(         j     )
+                + H3i       * XDATA(imax-1,j     )
+                +       H0j * XDATA(i     ,0     )
+                +       H1j * DXS(  i            )
+                +       H2j * DXN(  i            )
+                +       H3j * XDATA(i     ,jmax-1)
+                - H0i * H0j * XDATA(0     ,0     )
+                - H3i * H0j * XDATA(imax-1,0     )
+                - H0i * H3j * XDATA(0     ,jmax-1)
+                - H3i * H3j * XDATA(imax-1,jmax-1);
+            YDATA(i,j) =
+                + H0i       * YDATA(0     ,j     )
+                + H1i       * DYW(         j     )
+                + H2i       * DYE(         j     )
+                + H3i       * YDATA(imax-1,j     )
+                +       H0j * YDATA(i     ,0     )
+                +       H1j * DYS(  i            )
+                +       H2j * DYN(  i            )
+                +       H3j * YDATA(i     ,jmax-1)
+                - H0i * H0j * YDATA(0     ,0     )
+                - H3i * H0j * YDATA(imax-1,0     )
+                - H0i * H3j * YDATA(0     ,jmax-1)
+                - H3i * H3j * YDATA(imax-1,jmax-1);
+            ZDATA(i,j) =
+                + H0i       * ZDATA(0     ,j     )
+                + H1i       * DZW(         j     )
+                + H2i       * DZE(         j     )
+                + H3i       * ZDATA(imax-1,j     )
+                +       H0j * ZDATA(i     ,0     )
+                +       H1j * DZS(  i            )
+                +       H2j * DZN(  i            )
+                +       H3j * ZDATA(i     ,jmax-1)
+                - H0i * H0j * ZDATA(0     ,0     )
+                - H3i * H0j * ZDATA(imax-1,0     )
+                - H0i * H3j * ZDATA(0     ,jmax-1)
+                - H3i * H3j * ZDATA(imax-1,jmax-1);
         }
     }
 
-    FREE(xyztemp);
+    FREE(lens);
+    FREE(lenn);
+    FREE(lenw);
+    FREE(lene);
 
     status = EG_deleteObject(newSurface);
     CHECK_STATUS(EG_deleteObject);
@@ -709,6 +850,10 @@ cleanup:
 
     FREE(myEdges);
     FREE(dotprod);
+    FREE(lens);
+    FREE(lenn);
+    FREE(lenw);
+    FREE(lene);
 
     if (strlen(message) > 0) {
         MALLOC(*string, char, 1024);
